#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import copy
import heapq
import itertools
import sys
import traceback
from typing import Optional, cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
from scipy import spatial
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
import pyhope.output.output as hopout
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def flip_analytic(side: list[int], nbside: list[int]) -> int:
    """ Determines the flip of the side-to-side connection based on the analytic side ID
        flip = 1 : 1st node of neighbor side = 1st node of side
        flip = 2 : 2nd node of neighbor side = 1st node of side
        flip = 3 : 3rd node of neighbor side = 1st node of side
        flip = 4 : 4th node of neighbor side = 1st node of side
    """
    # Local imports ----------------------------------------
    from pyhope.common.common import find_index
    # ------------------------------------------------------
    return find_index(nbside, side[0])


def flip_physical(corners: np.ndarray, nbcorners: np.ndarray, tol: float, msg: str) -> int:
    """ Determines the flip of the side-to-side connection based on the physical positions
        flip = 1 : 1st node of neighbor side = 1st node of side
        flip = 2 : 2nd node of neighbor side = 1st node of side
        flip = 3 : 3rd node of neighbor side = 1st node of side
        flip = 4 : 4th node of neighbor side = 1st node of side
    """
    ptree     = spatial.KDTree(nbcorners)

    trCorn    = ptree.query(corners)
    if trCorn[0] > tol:
        hopout.warning(f'Could not determine flip of {msg} side within tolerance {tol}, exiting...')
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)

    flipID = cast(int, trCorn[1]) + 1
    return flipID


def connect_sides(sideIDs: list[int], sides: list, flipID: int) -> None:
    """ Connect the master and slave sides
    """
    # sides[sideIDs[0]].update(
    #     # Master side contains positive global side ID
    #     MS         = 1,                         # noqa: E251
    #     connection = sideIDs[1],                # noqa: E251
    #     flip       = flipID,                    # noqa: E251
    #     nbLocSide  = sides[sideIDs[1]].locSide  # noqa: E251
    # )
    # sides[sideIDs[1]].update(
    #     MS         = 0,                         # noqa: E251
    #     connection = sideIDs[0],                # noqa: E251
    #     flip       = flipID,                    # noqa: E251
    #     nbLocSide  = sides[sideIDs[0]].locSide  # noqa: E251
    # )
    sides[sideIDs[0]].MS         = 1                          # noqa: E251
    sides[sideIDs[0]].connection = sideIDs[1]                 # noqa: E251
    sides[sideIDs[0]].flip       = flipID                     # noqa: E251
    sides[sideIDs[0]].nbLocSide  = sides[sideIDs[1]].locSide  # noqa: E251
    sides[sideIDs[1]].MS         = 0                          # noqa: E251
    sides[sideIDs[1]].connection = sideIDs[0]                 # noqa: E251
    sides[sideIDs[1]].flip       = flipID                     # noqa: E251
    sides[sideIDs[1]].nbLocSide  = sides[sideIDs[0]].locSide  # noqa: E251


def connect_mortar_sides(sideIDs: list, elems: list, sides: list, nConnSide: list) -> list:
    """ Connect the master (big mortar) and the slave (small mortar) sides
        > Create the virtual sides as needed
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.mesh.mesh_vars import SIDE
    from pyhope.mesh.mesh_common import type_to_mortar_flip
    # ------------------------------------------------------

    # Get the master and slave sides
    masterSide    = sides[sideIDs[0]]
    masterElem    = elems[masterSide.elemID]
    # masterType    = masterElem['Type']
    masterCorners = masterSide.corners

    # Build mortar type and orientation
    nMortars = len(sideIDs[1])
    match nMortars:
        case 2:
            # Check which edges of big and small side are identical to determine the mortar type
            slaveSide    = sides[sideIDs[1][0]]
            slaveCorners = slaveSide.corners

            if   all(s in slaveCorners for s in [masterCorners[0], masterCorners[1]]) or \
                 all(s in slaveCorners for s in [masterCorners[2], masterCorners[3]]):  # noqa: E271
                mortarType = 2
            elif all(s in slaveCorners for s in [masterCorners[1], masterCorners[2]]) or \
                 all(s in slaveCorners for s in [masterCorners[0], masterCorners[3]]):
                mortarType = 3
            else:
                hopout.warning('Could not determine mortar type, exiting...')
                traceback.print_stack(file=sys.stdout)
                sys.exit(1)

            del slaveSide
            del slaveCorners

            # Sort the small sides
            # > Order according to master corners, [0, 2]
            slaveSides = [ sides[sideID] for i in [0, 2      ]
                                         for sideID in sideIDs[1] if masterCorners[i] in sides[sideID].corners]

        case 4:
            mortarType = 1
            # Sort the small sides
            # > Order according to master corners, [0, 1, 3, 2]
            slaveSides = [ sides[sideID] for i in [0, 1, 3, 2]
                                         for sideID in sideIDs[1] if masterCorners[i] in sides[sideID].corners]

        case _:
            hopout.warning('Found invalid number of sides for mortar side, exiting...')
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)

    # Update the master side
    # sides[sideIDs[0]].update(
    #     # Master side contains positive global side ID
    #     MS          = 1,            # noqa: E251
    #     connection  = -mortarType,  # noqa: E251
    #     flip        = 0,            # noqa: E251
    #     nbLocSide   = 0,            # noqa: E251
    # )
    sides[sideIDs[0]].MS          = 1            # noqa: E251
    sides[sideIDs[0]].connection  = -mortarType  # noqa: E251
    sides[sideIDs[0]].flip        = 0            # noqa: E251
    sides[sideIDs[0]].nbLocSide   = 0            # noqa: E251

    # Update the elems
    for elem in elems:
        for key, val in enumerate(elem.sides):
            # Update the sideIDs
            if val > masterSide.sideID:
                sides[val].sideID += nMortars
                elem.sides[key]   += nMortars
            # Update the connections
            if sides[val].connection is not None and sides[val].connection > masterSide.sideID:
                sides[val].connection += nMortars

    flipMap = type_to_mortar_flip(mesh_vars.elems[masterSide.elemID].type)

    match mortarType:
        case 1:  # 4-1 mortar
            mortarCorners = [0, 1, 3, 2]  # Prepare for non-quad mortars
        case 2:  # 2-1 mortar, split in eta
            mortarCorners = [0, 3]  # Prepare for non-quad mortars
        case 3:  # 2-1 mortar, split in xi
            mortarCorners = [0, 2]  # Prepare for non-quad mortars

    # Insert the virtual sides
    for key, val in enumerate(slaveSides):
        tol        = mesh_vars.tolInternal
        points     = mesh_vars.mesh.points[masterSide.corners]
        nbcorners  = mesh_vars.mesh.points[val.corners]

        flipID = flip_physical(points[mortarCorners[key]], nbcorners, tol, 'mortar')
        flipID = flipMap.get(mortarCorners[key], {}).get(flipID, flipID)
        # val.update(flip=flipID)
        val.flip = flipID

        # Insert the virtual sides
        side = SIDE(sideType   = 104,                          # noqa: E251
                    elemID     = masterElem.elemID,            # noqa: E251
                    sideID     = masterSide.sideID + key + 1,  # noqa: E251
                    locSide    = masterSide.locSide,           # noqa: E251
                    locMortar  = key + 1,                      # noqa: E251
                    # Virtual sides are always master sides
                    MS         = 1,                            # noqa: E251
                    flip       = flipID,                       # noqa: E251
                    connection = val.sideID,                   # noqa: E251
                    nbLocSide  = val.locSide                   # noqa: E251
                   )

        sides.insert(masterSide.sideID + key + 1, side)
        elems[masterElem.elemID].sides.insert(masterSide.locSide + key, side.sideID)

        # Connect the small (slave) sides to the master side
        # val.update(connection = masterSide.sideID,             # noqa: E251
        #            # Small sides are always slave sides
        #            sideType   = -104,                          # noqa: E251
        #            MS         = 0,                             # noqa: E251
        #            flip       = flipID,                        # noqa: E251
        #           )
        val.connection = masterSide.sideID             # noqa: E251
        val.sideType   = -104                          # noqa: E251
        val.MS         = 0                             # noqa: E251
        val.flip       = flipID                        # noqa: E251

        for s in nConnSide:
            if s.sideID == val.sideID:
                nConnSide.remove(s)
                break

    return nConnSide


def find_bc_index(bcs: list, key: str) -> Optional[int]:
    """ Find the index of a BC from its name in the list of BCs
    """
    for iBC, bc in enumerate(bcs):
        if key in bc.name:
            return iBC
        # Try again without the leading 'BC_'
        if key[:3] == 'BC_' and key[3:] in bc.name:
            return iBC
    return None


def find_closest_side(points: np.ndarray, stree: spatial.KDTree, tol: float, msg: str, doMortars: bool = False) -> int:
    """ Query the tree for the closest side
    """
    trSide = stree.query(points)

    # Check if the found side is within tolerance
    # trSide contains the Euclidean distance and the index of the
    # opposing side in the nbFaceSet
    if trSide[0] > tol:
        # Mortar sides are allowed to be not connected
        if doMortars:
            return -1

        hopout.warning(f'Could not find {msg} side within tolerance {tol}, exiting...')
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)
    return cast(int, trSide[1])


def find_mortar_match(targetCorners: np.ndarray, comboSides: list, mesh: meshio.Mesh, tol: float) -> bool:
    """ Check if the combined points of candidate sides match the target side within tolerance.
    """
    targetPoints = mesh.points[targetCorners]
    ttree = spatial.KDTree(targetPoints)

    comboCorners = [s.corners for s in comboSides]
    comboPoints  = np.concatenate([mesh.points[c] for c in comboCorners], axis=0)
    distances, indices = ttree.query(comboPoints)

    # At least one combo point must match each target point
    matchedIndices = np.unique(indices[distances <= tol])
    if len(matchedIndices) < 4:
        return False

    # Check if exactly one combo point matches each target point
    for point in targetPoints:
        # if not np.allclose(comboPoints, point, atol=tol, rtol=0):
        if np.sum(np.linalg.norm(comboPoints - point, axis=1) <= tol) != 1:
            return False

    # Build the target edges
    targetEdges = build_edges(targetCorners, mesh)
    matches     = []

    # First, check for 2-1 matches
    if len(comboSides) == 2:
        sideEdges = [build_edges(side.corners, mesh) for side in comboSides]

        # Look for 2-1 matches, we need exactly one common edge
        for edge in sideEdges[0]:
            targetP    = edge[:2]  # Start and end points (iX, jX)
            targetDist = edge[2]   # Distance between points

            # Initialize a list to store the matching combo edges for the current target edge
            matchEdges = []

            for comboEdge in sideEdges[1]:
                comboP    = comboEdge[:2]  # Start and end points (iX, jX)
                comboDist = comboEdge[2]   # Distance between points

                # Check if the points match and the distance is the same, taking into account the direction
                if (all(np.isclose(tp, cp) for tp, cp in zip(targetP, comboP)) or
                    all(np.isclose(tp, cp) for tp, cp in zip(targetP, reversed(comboP)))) and \
                    np.isclose(targetDist, comboDist):
                    matchEdges.append(comboEdge)

            # This should result in exactly 1 match
            if len(matchEdges) == 1:
                matches.append((edge, matchEdges.pop()))

        # We only allow 2-1 matches, so in the end we should have exactly 1 match
        if len(matches) != 1:
            return False

        # Here, we only allow 2-1 matches
        comboEdges  = [e for s in comboSides for e in build_edges(s.corners, mesh)]
        comboEdges  = find_edge_combinations(comboEdges)

        # Attempt to match the target edges with the candidate edges
        matches     = []  # List to store matching edges

        # Iterate over each target edge
        for targetEdge in targetEdges:
            targetP    = targetEdge[:2]  # Start and end points (iX, jX)
            targetDist = targetEdge[2]   # Distance between points

            # Initialize a list to store the matching combo edges for the current target edge
            matchEdges = []

            # Iterate over comboEdges to find matching edges
            for comboEdge in comboEdges:
                comboP    = comboEdge[:2]  # Start and end points (iX, jX)
                comboDist = comboEdge[2]   # Distance between points

                # Check if the points match and the distance is the same, taking into account the direction
                if (all(np.isclose(tp, cp) for tp, cp in zip(targetP, comboP)) or
                    all(np.isclose(tp, cp) for tp, cp in zip(targetP, reversed(comboP)))) and \
                   np.isclose(targetDist, comboDist):
                    matchEdges.append(comboEdge)

            # This should result in exactly 1 match
            if len(matchEdges) > 1:
                return False
            elif len(matchEdges) == 1:
                matches.append((targetEdge, matchEdges.pop()))

        if len(matches) != 2:
            return False

    # Next, check for 4-1 matches
    if len(comboSides) == 4:
        # Check if there is exactly one point that all 4 sides have in common.
        common_points = set(comboSides[0].corners)
        matchFound = False
        for p in common_points:
            # Check if all 4 sides have the point
            matchedPoints = 0
            for side in comboSides[1:]:
                for p1 in side.corners:
                    if (all(np.isclose(tp, cp) for tp, cp in zip(mesh.points[p], mesh.points[p1]))):
                        matchedPoints += 1

            if matchedPoints == 3:
                matchFound = True
                break

        if not matchFound:
            return False

        comboEdges  = [e for s in comboSides for e in build_edges(s.corners, mesh)]
        comboEdges  = find_edge_combinations(comboEdges)

        # Attempt to match the target edges with the candidate edges
        matches     = []  # List to store matching edges

        # Iterate over each target edge
        for targetEdge in targetEdges:
            targetP    = targetEdge[:2]  # Start and end points (iX, jX)
            targetDist = targetEdge[2]   # Distance between points

            # Initialize a list to store the matching combo edges for the current target edge
            matchEdges = []

            # Iterate over comboEdges to find matching edges
            for comboEdge in comboEdges:
                comboP    = comboEdge[:2]  # Start and end points (iX, jX)
                comboDist = comboEdge[2]  # Distance between points

                # Check if the points match and the distance is the same, taking into account the direction
                if (all(np.isclose(tp, cp) for tp, cp in zip(targetP, comboP)) or
                    all(np.isclose(tp, cp) for tp, cp in zip(targetP, reversed(comboP)))) and \
                   np.isclose(targetDist, comboDist):
                    matchEdges.append(comboEdge)

            # This should result in exactly 1 match
            if len(matchEdges) > 1:
                return False
            elif len(matchEdges) == 1:
                matches.append((targetEdge, matchEdges.pop()))

        if len(matches) != 4:
            return False

    # Found a valid match
    return True


def build_edges(corners: np.ndarray, mesh: meshio.Mesh) -> list[tuple]:
    """Build edges from the 4 corners of a quadrilateral, considering CGNS ordering
    """
    edges = [
        (corners[0], corners[1], np.linalg.norm(mesh.points[corners[0]] - mesh.points[corners[1]])),  # Edge between points 0 and 1
        (corners[1], corners[2], np.linalg.norm(mesh.points[corners[1]] - mesh.points[corners[2]])),  # Edge between points 1 and 2
        (corners[2], corners[3], np.linalg.norm(mesh.points[corners[2]] - mesh.points[corners[3]])),  # Edge between points 2 and 3
        (corners[3], corners[0], np.linalg.norm(mesh.points[corners[3]] - mesh.points[corners[0]])),  # Edge between points 3 and 0
    ]
    return edges


def find_edge_combinations(comboEdges):
    """Build combinations of edges that share exactly one point and form a line
    """
    # Local imports ----------------------------------------
    from collections import defaultdict
    from itertools import combinations
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------

    # Create a dictionary to store edges by their shared points
    pointToEdges = defaultdict(list)

    # Fill the dictionary with edges indexed by their points
    for i, j, dist in comboEdges:
        pointToEdges[i].append((i, j, dist))
        pointToEdges[j].append((i, j, dist))

    # Initialize an empty list to store the valid combinations of edges
    validCombo = []

    # Iterate over all points and their associated edges
    for _, edges in pointToEdges.items():
        if len(edges) < 2:  # Skip points with less than 2 edges
            continue

        # Now, we generate all possible pairs of edges that share the point
        for edge1, edge2 in combinations(edges, 2):
            # Ensure the edges are distinct and share exactly one point
            # Since both edges share 'point', they are valid combinations
            # We store the combination as an np.array (i, j, dist)
            i1, j1, _ = edge1
            i2, j2, _ = edge2

            # Use set operations to determine the unique start and end points
            commonPoint = {i1, j1} & {i2, j2}
            if len(commonPoint) == 1:  # Check that there's exactly one shared point
                commonPoint = commonPoint.pop()

                # Exclude the common point and get the unique start and end points
                edgePoints = np.array([i1, j1, i2, j2])

                # Find the index of the common point and delete it
                commonIndex = np.where( edgePoints == commonPoint)[0]
                edgePoints  = np.delete(edgePoints, commonIndex)

                # The remaining points are the start and end points of the edge combination
                point1, point2 = edgePoints

                # Get the coordinates of the points
                p1, p2 = mesh_vars.mesh.points[point1], mesh_vars.mesh.points[point2]
                c1     = mesh_vars.mesh.points[commonPoint]

                # Calculate the bounding box of the two edge points
                bbox_min = np.minimum(p1, p2)
                bbox_max = np.maximum(p1, p2)

                # Check if the common point is within the bounding box of p1 and p2
                if np.all(np.isclose(bbox_min, np.minimum(bbox_min, c1))) and \
                   np.all(np.isclose(bbox_max, np.maximum(bbox_max, c1))):
                    # Calculate the distance between the start and end points
                    lineDist = np.linalg.norm(p1 - p2)

                    # Append the indices and the line distance
                    validCombo.append((point1, point2, lineDist))

    return validCombo


def get_side_id(corners: np.ndarray, side_dict: dict) -> int:
    """ Get sorted corners and hash them to get the side ID
    """
    corners_sorted = np.sort(corners)
    corners_hash = hash(corners_sorted.tobytes())
    return side_dict[corners_hash][0]


def get_nonconnected_sides(sides: list, mesh: meshio.Mesh) -> tuple[list, list[np.ndarray]]:
    """ Get a list of internal sides that are not connected to any
        other side together with a list of their centers
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------
    # Update the list
    nConnSide = [s for s in sides if s.connection is None and s.bcid is None]
    # Append the inner BCs
    for s in (s for s in sides if s.bcid is not None and s.connection is None):
        if mesh_vars.bcs[s.bcid].type[0] == 0:
            nConnSide.append(s)
    nConnCenter = [np.mean(mesh.points[s.corners], axis=0) for s in nConnSide]
    return nConnSide, nConnCenter


def periodic_update(sideIDs: list[int], vv: dict) -> None:
    """Update the mesh after connecting periodic sides
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.mesh.mesh_common import face_to_nodes
    from pyhope.mesh.mesh_common import flip_s2m
    # ------------------------------------------------------

    sides    = [mesh_vars.sides[s] for s in sideIDs]
    elems    = [mesh_vars.elems[s.elemID] for s in sides]
    nodes    = np.array([elems[0].nodes[s] for s in face_to_nodes(sides[0].face, elems[0].type, mesh_vars.nGeo)])
    nbNodes  = np.array([elems[1].nodes[s] for s in face_to_nodes(sides[1].face, elems[1].type, mesh_vars.nGeo)])

    # Get the flip map
    indices = flip_s2m(mesh_vars.nGeo+1, 1 if sides[0].flip <= 2 else sides[0].flip)

    # for iy, ix in np.ndindex(nodes.shape[:2]):
    #     node   = nodes[ix, iy]
    #     nbNode = nbNodes[indices[ix, iy, 0], indices[ix, iy, 1]]
    #
    #     # Sanity check if the periodic vector matches
    #     if not np.allclose(vv['Dir'], mesh_vars.mesh.points[nbNode] - mesh_vars.mesh.points[node],
    #                        rtol=mesh_vars.tolPeriodic, atol=mesh_vars.tolPeriodic):
    #         hopout.warning('Error in periodic update, periodic vector does not match!')
    #         sys.exit(1)
    #
    #     # Center between both points
    #     center = 0.5 * (mesh_vars.mesh.points[node] + mesh_vars.mesh.points[nbNode])
    #
    #     lowerP = copy.copy(center)
    #     upperP = copy.copy(center)
    #     for key, val in enumerate(vv['Dir']):
    #         lowerP[key] -= 0.5 * val
    #         upperP[key] += 0.5 * val
    #
    #     mesh_vars.mesh.points[  node] = lowerP
    #     mesh_vars.mesh.points[nbNode] = upperP

    # Extract relevant indices from the mesh
    nbNodes = nbNodes[indices[:, :, 0], indices[:, :, 1]]

    # Check if periodic vector matches using vectorized np.allclose
    if not np.allclose(mesh_vars.mesh.points[nodes] + vv['Dir'], mesh_vars.mesh.points[nbNodes],
                       rtol=mesh_vars.tolPeriodic, atol=mesh_vars.tolPeriodic):
        hopout.warning('Error in periodic update, periodic vector does not match!')
        sys.exit(1)

    # Calculate the center for both points
    centers = 0.5 * (mesh_vars.mesh.points[nodes] + mesh_vars.mesh.points[nbNodes])

    # Update the mesh points for both node and nbNode
    mesh_vars.mesh.points[nodes]   = centers - 0.5*vv['Dir']
    mesh_vars.mesh.points[nbNodes] = centers + 0.5*vv['Dir']


def ConnectMesh() -> None:
    # Local imports ----------------------------------------
    import pyhope.io.io_vars as io_vars
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.common.common import find_index
    from pyhope.common.common_progress import ProgressBar
    from pyhope.io.io_vars import MeshFormat
    from pyhope.readintools.readintools import GetLogical
    from pyhope.mesh.mesh_common import face_to_nodes
    # ------------------------------------------------------

    match io_vars.outputformat:
        case MeshFormat.FORMAT_HDF5:
            pass
        case _:
            return

    hopout.separator()
    hopout.info('CONNECT MESH...')
    hopout.sep()

    mesh_vars.doPeriodicCorrect = GetLogical('doPeriodicCorrect')
    mesh_vars.doMortars         = GetLogical('doMortars')
    doPeriodicCorrect = mesh_vars.doPeriodicCorrect
    doMortars         = mesh_vars.doMortars

    mesh    = mesh_vars.mesh
    elems   = mesh_vars.elems
    sides   = mesh_vars.sides

    # cell_sets contain the face IDs [dim=2]
    # > Offset is calculated with entities from [dim=0, dim=1]
    offsetcs = 0
    for key, value in mesh.cells_dict.items():
        if 'vertex' in key:
            offsetcs += value.shape[0]
        elif 'line' in key:
            offsetcs += value.shape[0]

    # Map sides to BC
    # > Create a dict containing only the face corners
    side_corners = dict()
    for elem in elems:
        for iSide, side in enumerate(elem.sides):
            corners = np.sort(sides[side].corners)
            corners = hash(corners.tobytes())
            side_corners.update({side: corners})

    # Build the reverse dictionary
    corner_side = dict()
    for key, val in side_corners.items():
        if val not in corner_side:
            corner_side[val] = [key]
        else:
            corner_side[val].append(key)

    bar = ProgressBar(value=len(sides), title='│                Processing Sides')

    # Try to connect the inner sides
    ninner = 0
    for (key, val) in corner_side.items():
        match len(val):
            case 1:  # BC side
                continue
            case 2:  # Internal side
                sideIDs   = val
                corners   = sides[sideIDs[0]].corners
                nbcorners = sides[sideIDs[1]].corners
                flipID    = flip_analytic(corners, nbcorners) + 1
                # Connect the sides
                connect_sides(sideIDs, sides, flipID)
                ninner += 1

                # Update the progress bar
                [bar.step() for _ in range(2)]
            case _:  # Zero or more than 2 sides
                hopout.warning('Found internal side with more than two adjacent elements, exiting...')
                traceback.print_stack(file=sys.stdout)
                sys.exit(1)

    # Set BC and periodic sides
    bcs = mesh_vars.bcs
    vvs = mesh_vars.vvs
    csetMap = []
    for key, cset in mesh.cell_sets.items():
        # Check if the set is a BC
        bcID = find_bc_index(bcs, key)
        if bcID is None:
            hopout.warning(f'Could not find BC {key} in list, exiting...')
            sys.exit(1)

        # Find the mapping to the (N-1)-dim elements
        csetMap = [s for s in range(len(cset)) if cset[s] is not None]

        # Get the list of sides
        for iMap in csetMap:
            iBCsides = np.array(cset[iMap]).astype(int) - offsetcs
            mapFaces = mesh.cells[iMap].data

            # Map the unique quad sides to our non-unique elem sides
            for iSide in iBCsides:
                # Get the quad corner nodes
                # FIXME: HARDCODED FIRST 4 NODES WHICH ARE THE OUTER CORNER NODES FOR QUADS!
                corners = np.sort(np.array(mapFaces[iSide][0:4]))
                corners = hash(corners.tobytes())

                # Boundary faces are unique
                # sideID  = find_key(face_corners, corners)
                sideID = corner_side[corners][0]
                # sides[sideID].update(bcid=bcID)
                sides[sideID].bcid = bcID

                if bcs[bcID].type[0] != 1:
                    bar.step()

    # Try to connect the periodic sides
    for key, cset in mesh.cell_sets.items():
        # Check if the set is a BC
        bcID = find_bc_index(bcs, key)
        if bcID is None:
            hopout.warning(f'Could not find BC {key} in list, exiting...')
            sys.exit(1)

        # Only periodic BCs and only try to connect in positive direction
        if bcs[bcID].type[0] != 1 or bcs[bcID].type[3] < 0:
            continue

        # Get the opposite side
        iVV        = bcs[bcID].type[3] - 1
        nbType     = copy.copy(bcs[bcID].type)
        nbType[3] *= -1
        nbBCID     = find_index([s.type for s in bcs], nbType)
        nbBCName   = bcs[nbBCID].name

        # Collapse all opposing corner nodes into an [:, 12] array
        nbCellSet  = mesh.cell_sets[nbBCName]
        # Find the mapping to the (N-1)-dim elements
        nbcsetMap  = [s for s in range(len(nbCellSet)) if nbCellSet[s] is not None
                      and cast(np.ndarray, nbCellSet[s]).size > 0]

        # FIXME: TODO HYBRID MESHES
        if len(nbcsetMap) > 1:
            print('Hybrid meshes currently not supported')
            sys.exit(1)

        # Get the list of sides
        nbFaceSet  =  np.array(nbCellSet[csetMap[0]]).astype(int)
        nbmapFaces = mesh.cells[csetMap[0]].data
        nbCorners  = [np.array(nbmapFaces[s - offsetcs]) for s in nbFaceSet]
        nbPoints   = copy.copy(np.sort(mesh.points[nbCorners], axis=1))
        nbPoints   = nbPoints.reshape(nbPoints.shape[0], nbPoints.shape[1]*nbPoints.shape[2])
        del nbCorners

        # Build a k-dimensional tree of all points on the opposing side
        stree    = spatial.KDTree(nbPoints)
        # Get the list of sides on our side
        iBCsides = np.array(cset[csetMap[0]]).astype(int) - offsetcs

        # Map the unique quad sides to our non-unique elem sides
        for iSide in iBCsides:
            # Get the quad corner nodes
            corners   = np.array(nbmapFaces[iSide])
            points    = copy.copy(mesh.points[corners])

            # Shift the points in periodic direction
            for iPoint in range(points.shape[0]):
                points[iPoint, :] += vvs[iVV]['Dir']
            points    = np.sort(points, axis=0).flatten()

            # Query the try for the opposing side
            tol       = np.linalg.norm(vvs[iVV]['Dir'], ord=2).astype(float) * mesh_vars.tolPeriodic
            nbSideIdx = find_closest_side(points, cast(spatial.KDTree, stree), tol, 'periodic')
            nbiSide   = nbFaceSet[nbSideIdx] - offsetcs

            # Get our and neighbor corner quad nodes
            sideID    = get_side_id(nbmapFaces[iSide  ][0:4], corner_side)
            nbSideID  = get_side_id(nbmapFaces[nbiSide][0:4], corner_side)

            # Build the connection, including flip
            sideIDs   = [sideID, nbSideID]
            points    = mesh.points[sides[sideIDs[0]].corners]
            for iPoint in range(points.shape[0]):
                points[iPoint, :] += vvs[iVV]['Dir']

            # > Find the first neighbor point to determine the flip
            nbcorners = mesh.points[sides[sideIDs[1]].corners]
            flipID    = flip_physical(points[0], nbcorners, tol, 'periodic')

            # Connect the sides
            connect_sides(sideIDs, sides, flipID)

            # Update the sides
            if doPeriodicCorrect:
                periodic_update(sideIDs, vvs[iVV])

                # Update the progress bar
                [bar.step() for _ in range(2)]

    # Non-connected sides without BCID are possible inner sides
    nConnSide, nConnCenter = get_nonconnected_sides(sides, mesh)
    nInterZoneConnect      = len(nConnSide)

    # Loop over all sides and try to connect
    iter    = 0
    maxIter = copy.copy(nInterZoneConnect)
    tol     = mesh_vars.tolInternal
    # While maxIter should be enough, this results in non-connected mortar sides. We can append a maximum of 4 virtual sides,
    # so let's set the maxIter to 5 just to be safe.
    while len(nConnSide) > 1 and iter <= maxIter:
        # Ensure the loop exits after checking every side
        iter += 1

        # Remove the first side from the list
        targetSide   = nConnSide  .pop(0)
        targetCenter = nConnCenter.pop(0)

        # Collapse all opposing corner nodes into an [:, 12] array
        nbCorners  = [s.corners for s in nConnSide]
        nbPoints   = copy.copy(np.sort(mesh.points[nbCorners], axis=1))
        nbPoints   = nbPoints.reshape(nbPoints.shape[0], nbPoints.shape[1]*nbPoints.shape[2])
        del nbCorners

        # Build a k-dimensional tree of all points on the opposing side
        stree      = spatial.KDTree(nbPoints)
        ctree      = spatial.KDTree(nConnCenter)

        # Map the unique quad sides to our non-unique elem sides
        corners    = targetSide.corners
        points     = np.sort(mesh.points[corners], axis=0).flatten()

        # Query the tree for the opposing side
        nbSideIdx  = find_closest_side(points, cast(spatial.KDTree, stree), tol, 'internal', doMortars)

        # Regular internal side
        if nbSideIdx >= 0:
            nbiSide   = nbSideIdx

            # Get our and neighbor corner quad nodes
            sideID    = get_side_id(targetSide.corners        , corner_side)
            nbSideID  = get_side_id(nConnSide[nbiSide].corners, corner_side)

            # Build the connection, including flip
            sideIDs   = [sideID, nbSideID]
            points    = mesh.points[sides[sideIDs[0]].corners]
            # > Find the first neighbor point to determine the flip
            nbcorners = mesh.points[sides[sideIDs[1]].corners]
            flipID    = flip_physical(points[0], nbcorners, tol, 'internal')

            # Connect the sides
            connect_sides(sideIDs, sides, flipID)

            # Update the list
            nConnSide, nConnCenter = get_nonconnected_sides(sides, mesh)

            # Update the progress bar
            [bar.step() for _ in range(2)]

            if not doMortars:
                hopout.warning(f'Could not find internal side within tolerance {tol}, exiting...')
                traceback.print_stack(file=sys.stdout)
                sys.exit(1)

    # Mortar sides
    if doMortars:
        nConnSide, nConnCenter = get_nonconnected_sides(sides, mesh)
        nInterZoneConnect      = len(nConnSide)

        iter    = 0
        maxIter = copy.copy(nInterZoneConnect)
        tol     = mesh_vars.tolInternal

        while len(nConnSide) > 1 and iter <= maxIter:
            # Ensure the loop exits after checking every side
            iter += 1

            # Remove the first side from the list
            targetSide   = nConnSide  .pop(0)
            targetCenter = nConnCenter.pop(0)

            # Collapse all opposing corner nodes into an [:, 12] array
            nbCorners  = [s.corners for s in nConnSide]
            nbPoints   = copy.copy(np.sort(mesh.points[nbCorners], axis=1))
            nbPoints   = nbPoints.reshape(nbPoints.shape[0], nbPoints.shape[1]*nbPoints.shape[2])
            del nbCorners

            # Build a k-dimensional tree of all points on the opposing side
            stree      = spatial.KDTree(nbPoints)
            ctree      = spatial.KDTree(nConnCenter)

            # Map the unique quad sides to our non-unique elem sides
            corners    = targetSide.corners
            points     = np.sort(mesh.points[corners], axis=0).flatten()

            # Query the tree for the opposing side
            nbSideIdx  = find_closest_side(points, cast(spatial.KDTree, stree), tol, 'internal', doMortars)

            # Mortar side
            # > Here, we can only attempt to connect big to small mortar sides. Thus, if we encounter a small mortar sides which
            # > generates no match, we simply append the side again at the end and try again. As the loop exists after checking
            # > len(nConnSide), we will check each side once.
            if nbSideIdx >= 0:
                continue

            # Calculate the radius of the convex hull
            targetPoints = mesh.points[targetSide.corners]
            targetMinMax = (targetPoints.min(axis=0), targetPoints.max(axis=0))
            targetRadius = np.linalg.norm(targetMinMax[1] - targetMinMax[0], ord=2) / 2.

            # Find nearby sides to consider as candidate mortar sides
            targetNeighbors = ctree.query_ball_point(targetCenter, targetRadius)
            # Eliminate sides in the same element
            targetNeighbors = [s for s in targetNeighbors if nConnSide[s].elemID != targetSide.elemID]

            # Prepare combinations for 2-to-1 and 4-to-1 mortar matching
            candidate_combinations = []
            if len(targetNeighbors) >= 2:
                candidate_combinations += list(itertools.combinations(targetNeighbors, 2))
            if len(targetNeighbors) >= 4:
                candidate_combinations += list(itertools.combinations(targetNeighbors, 4))

            # Attempt to match the target side with candidate combinations
            matchFound   = False
            comboSides   = []
            for comboIDs in candidate_combinations:
                # Get the candidate sides
                comboSides   = [nConnSide[iSide] for iSide in comboIDs]

                # Found a valid match
                if find_mortar_match(targetSide.corners, comboSides, mesh, tol):
                    matchFound = True
                    break

            if matchFound:
                # Get our and neighbor corner quad nodes
                sideID    =  targetSide.sideID
                nbSideID  = [side.sideID for side in comboSides]

                # Build the connection, including flip
                sideIDs   = [sideID, nbSideID]

                # Connect mortar sides and update the list
                nConnSide   = connect_mortar_sides(sideIDs, elems, sides, nConnSide)
                nConnCenter = [np.mean(mesh.points[s.corners], axis=0) for s in nConnSide]

                # Update the progress bar
                [bar.step() for _ in range(len(nbSideID) + 1)]

            # No connection, attach the side at the end
            else:
                nConnSide  .append(targetSide)
                nConnCenter.append(targetCenter)

    nConnSide, nConnCenter = get_nonconnected_sides(sides, mesh)
    if len(nConnSide) > 0:
        hopout.warning('Could not connect {} side{}'.format(len(nConnSide), '' if len(nConnSide) == 1 else 's'))

        for side in nConnSide:
            print(hopout.warn(f'> Element {side.elemID+1}, Side {side.face}, Side {side.sideID+1}'))  # noqa: E501
            elem     = elems[side.elemID]
            nodes    = np.transpose(np.array([elem.nodes[s] for s in face_to_nodes(side.face, elem.type, mesh_vars.nGeo)]))
            nodes    = np.transpose(mesh_vars.mesh.points[nodes]         , axes=(2, 0, 1))
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nodes[:,  0,  0]) + ']'))
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nodes[:,  0, -1]) + ']'))
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nodes[:, -1,  0]) + ']'))
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nodes[:, -1, -1]) + ']'))
            print()
        sys.exit(1)

    # Close the progress bar
    bar.close()

    hopout.separator()
    if nInterZoneConnect > 0:
        hopout.sep()
        hopout.routine('Connected {} inter-zone faces'.format(nInterZoneConnect))

    # Set the global side ID
    globalSideID     = 0
    highestSideID    = 0
    usedSideIDs      = set()  # Set to track used side IDs
    availableSideIDs = []     # Min-heap for gap

    for iSide, side in enumerate(sides):
        # Already counted the side
        if side.globalSideID is not None:
            continue

        # Get the smallest available globalSideID from the heap, if any
        if availableSideIDs:
            globalSideID = heapq.heappop(availableSideIDs)
        else:
            # Use the current maximum ID and increment
            globalSideID = highestSideID + 1

        # Mark the side ID as used
        highestSideID = max(globalSideID, highestSideID)
        usedSideIDs.add(globalSideID)
        # side.update(globalSideID=globalSideID)
        side.globalSideID = globalSideID

        if side.connection is None:         # BC side
            pass
        elif side.connection < 0:           # Big mortar side
            pass
        elif side.MS == 1:                  # Internal / periodic side (master side)
            # Get the connected slave side
            nbSideID = side.connection

            # Reclaim the ID of the slave side if already assigned
            if sides[nbSideID].globalSideID is not None:
                reclaimedID = sides[nbSideID].globalSideID
                usedSideIDs.remove(reclaimedID)
                heapq.heappush(availableSideIDs, reclaimedID)

            # Set the negative globalSideID of the slave  side
            # sides[nbSideID].update(globalSideID=-(globalSideID))
            sides[nbSideID].globalSideID = -(globalSideID)

    # Count the sides
    nsides             = len(sides)
    sides_conn         = np.array([s.connection is not None                      for s in sides])  # noqa: E271, E272
    sides_bc           = np.array([s.bcid       is not None                      for s in sides])  # noqa: E271, E272
    sides_mortar_big   = np.array([s.connection is not None and s.connection < 0 for s in sides])  # noqa: E271, E272
    sides_mortar_small = np.array([s.locMortar  is not None                      for s in sides])  # noqa: E271, E272

    # Count each type of side
    ninnersides        = np.sum( sides_conn & ~sides_bc & ~sides_mortar_small & ~sides_mortar_big)
    nperiodicsides     = np.sum( sides_conn &  sides_bc & ~sides_mortar_small & ~sides_mortar_big)
    nbcsides           = np.sum(~sides_conn &  sides_bc & ~sides_mortar_small & ~sides_mortar_big)
    nmortarbigsides    = np.sum(                                                 sides_mortar_big)
    nmortarsmallsides  = np.sum(                           sides_mortar_small                    )
    nsides             = len(sides) - nmortarsmallsides

    hopout.sep()
    hopout.info(' Number of sides                : {:12d}'.format(nsides))
    hopout.info(' Number of inner sides          : {:12d}'.format(ninnersides))
    hopout.info(' Number of mortar sides (big)   : {:12d}'.format(nmortarbigsides))
    hopout.info(' Number of mortar sides (small) : {:12d}'.format(nmortarsmallsides))
    hopout.info(' Number of boundary sides       : {:12d}'.format(nbcsides))
    hopout.info(' Number of periodic sides       : {:12d}'.format(nperiodicsides))
    hopout.sep()

    hopout.info('CONNECT MESH DONE!')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
from functools import cache
from typing import Union
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
import numpy.typing as npt
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
import pyhope.mesh.mesh_vars as mesh_vars
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


@cache
def faces(elemType: Union[int, str]) -> list[str]:
    """ Return a list of all sides of a hexahedron
    """
    faces_map = {  # Tetrahedron
                   # Pyramid
                   # Wedge / Prism
                   # Hexahedron
                   8: ['z-', 'y-', 'x+', 'y+', 'x-', 'z+']
                }

    if isinstance(elemType, str):
        elemType = mesh_vars.ELEMTYPE.name[elemType]

    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in faces: elemType {elemType} is not supported')

    return faces_map[elemType % 100]


@cache
def edge_to_dir(edge: int, elemType: Union[int, str]) -> int:
    """ GMSH: Create edges from points in the given direction
    """
    dir_map  = {  # Tetrahedron
                  # Pyramid
                  # Wedge / Prism
                  # Hexahedron
                  8: {  0:  0,  2:  0,  4:  0,  6:  0,  # Direction 0
                        1:  1,  3:  1,  5:  1,  7:  1,  # Direction 1
                        8:  2,  9:  2, 10:  2, 11:  2}  # Direction 2
               }

    if isinstance(elemType, str):
        elemType = mesh_vars.ELEMTYPE.name[elemType]

    if elemType % 100 not in dir_map:
        raise ValueError(f'Error in edge_to_direction: elemType {elemType} is not supported')

    dir = dir_map[elemType % 100]

    try:
        return dir[edge]
    except KeyError:
        raise KeyError(f'Error in edge_to_dir: edge {edge} is not supported')


@cache
def edge_to_corner(edge: int, elemType: Union[int, str], dtype=int) -> np.ndarray:
    """ GMSH: Get points on edges
    """
    edge_map = {  # Tetrahedron
                  4: [ [0, 1], [1, 2], [2, 1], [0, 3],
                       [1, 3], [2, 3]                 ],
                  # Pyramid
                  5: [ [0, 1], [1, 2], [2, 3], [3, 0],
                       [0, 4], [1, 5], [2, 4], [3, 4] ],
                  # Wedge / Prism
                  6: [ [0, 1], [1, 2], [2, 0], [0, 3],
                       [2, 3], [3, 4], [4, 5], [5, 4] ],
                  # Hexahedron
                  8: [ [0, 1], [1, 2], [2, 3], [3, 0],
                       [0, 4], [1, 5], [2, 6], [3, 7],
                       [4, 5], [5, 6], [6, 7], [7, 4] ],
               }

    if isinstance(elemType, str):
        elemType = mesh_vars.ELEMTYPE.name[elemType]

    if elemType % 100 not in edge_map:
        raise ValueError(f'Error in edge_to_corner: elemType {elemType} is not supported')

    edges = edge_map[elemType % 100]

    try:
        return np.array(edges[edge], dtype=dtype)
    except KeyError:
        raise KeyError(f'Error in edge_to_corner: edge {edge} is not supported')


@cache
def face_to_edge(face: str, elemType: Union[str, int], dtype=int) -> np.ndarray:
    """ GMSH: Create faces from edges in the given direction
    """
    faces_map = {  # Tetrahedron
                   # Pyramid
                   # Wedge / Prism
                   # Hexahedron
                   8: {  'z-': np.array([  0,  1,   2,   3], dtype=dtype),
                         'y-': np.array([  0,  9,  -4,  -8], dtype=dtype),
                         'x+': np.array([  1, 10,  -5,  -9], dtype=dtype),
                         'y+': np.array([ -2, 10,   6, -11], dtype=dtype),
                         'x-': np.array([  8, -7, -11,   3], dtype=dtype),
                         'z+': np.array([  4,  5,   6,   7], dtype=dtype)}
                }

    if isinstance(elemType, str):
        elemType = mesh_vars.ELEMTYPE.name[elemType]

    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in face_to_edge: elemType {elemType} is not supported')

    try:
        return faces_map[elemType % 100][face]
    except KeyError:
        raise KeyError(f'Error in face_to_edge: face {face} is not supported')


@cache
def face_to_corner(face, elemType: Union[str, int], dtype=int) -> np.ndarray:
    """ GMSH: Get points on faces in the given direction
    """
    faces_map = {  # Tetrahedron
                   # Pyramid
                   # Wedge / Prism
                   # Hexahedron
                   8: {  'z-': np.array([  0,  1,   2,   3], dtype=dtype),
                         'y-': np.array([  0,  1,   5,   4], dtype=dtype),
                         'x+': np.array([  1,  2,   6,   5], dtype=dtype),
                         'y+': np.array([  2,  6,   7,   3], dtype=dtype),
                         'x-': np.array([  0,  4,   7,   3], dtype=dtype),
                         'z+': np.array([  4,  5,   6,   7], dtype=dtype)}
                }

    if isinstance(elemType, str):
        elemType = mesh_vars.ELEMTYPE.name[elemType]

    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in face_to_corner: elemType {elemType} is not supported')

    try:
        return faces_map[elemType % 100][face]
    except KeyError:
        raise KeyError(f'Error in face_to_corner: face {face} is not supported')


@cache
def face_to_cgns(face: str, elemType: Union[str, int], dtype=int) -> np.ndarray:
    """ CGNS: Get points on faces in the given direction
    """
    faces_map = {  # Tetrahedron
                   # Pyramid
                   # Wedge / Prism
                   # Hexahedron
                   8: {'z-': np.array([  0,  3,  2,  1], dtype=dtype),
                       'y-': np.array([  0,  1,  5,  4], dtype=dtype),
                       'x+': np.array([  1,  2,  6,  5], dtype=dtype),
                       'y+': np.array([  2,  3,  7,  6], dtype=dtype),
                       'x-': np.array([  0,  4,  7,  3], dtype=dtype),
                       'z+': np.array([  4,  5,  6,  7], dtype=dtype)}
                }

    if isinstance(elemType, str):
        elemType = mesh_vars.ELEMTYPE.name[elemType]

    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in face_to_cgns: elemType {elemType} is not supported')

    try:
        return faces_map[elemType % 100][face]
    except KeyError:
        raise KeyError(f'Error in face_to_cgns: face {face} is not supported')


@cache
def flip_s2m(N: int, flip: int) -> np.ndarray:
    # Create grid index arrays for the rows and columns
    p = np.arange(N)
    q = np.arange(N)

    # Create a meshgrid of row (p) and column (q) indices
    p_grid, q_grid = np.meshgrid(p, q)

    # Map row and column indices based on flip logic
    # WARNING: FOR SOME REASON, ONLY FLIP 1,3,4 IS USED WITH FACE_TO_NODES
    if flip == 0:
        return np.stack((q_grid        , p_grid        ), axis=-1)
    elif flip == 1:
        return np.stack((p_grid        , q_grid        ), axis=-1)
    elif flip == 2:
        return np.stack((N - p_grid - 1, q_grid        ), axis=-1)
    elif flip == 3:
        return np.stack((N - p_grid - 1, N - q_grid - 1), axis=-1)
    elif flip == 4:
        return np.stack((N - q_grid - 1, p_grid        ), axis=-1)
    else:
        raise ValueError('Flip must be an integer between 0 and 4')


@cache
def type_to_mortar_flip(elemType: Union[int, str]) -> dict[int, dict[int, int]]:
    """ Returns the flip map for a given element type
    """

    flipID_map = {  # Tetrahedron
                   # Pyramid
                   # Wedge / Prism
                   # Hexahedron
                   8: { 0: {1: 1, 2: 2, 3: 3, 4: 4},
                        1: {1: 2, 4: 1, 3: 4, 2: 3},
                        2: {3: 1, 4: 2, 1: 3, 2: 4},
                        3: {2: 1, 3: 2, 4: 3, 1: 4}}
                }

    if isinstance(elemType, str):
        elemType = mesh_vars.ELEMTYPE.name[elemType]

    if elemType % 100 not in flipID_map:
        raise ValueError(f'Error in type_to_mortar_flip: elemType {elemType} is not supported')

    try:
        return flipID_map[elemType % 100]
    except KeyError:
        raise KeyError(f'Error in type_to_mortar_flip: elemType {elemType} is not supported')


@cache
def face_to_nodes(face: str, elemType: int, nGeo: int) -> np.ndarray:
    """ Returns the tensor-product nodes associated with a face
    """
    if isinstance(elemType, str):
        elemType = mesh_vars.ELEMTYPE.name[elemType]

    order     = nGeo
    faces_map = {  # Tetrahedron
                   # Pyramid
                   # Wedge / Prism
                   # Hexahedron
                   8: { 'z-':              LINMAP(elemType, order=order)[:    , :    , 0    ],
                        'y-': np.transpose(LINMAP(elemType, order=order)[:    , 0    , :    ]),
                        'x+': np.transpose(LINMAP(elemType, order=order)[order, :    , :    ]),
                        'y+':              LINMAP(elemType, order=order)[:    , order, :    ],
                        'x-':              LINMAP(elemType, order=order)[0    , :    , :    ],
                        'z+': np.transpose(LINMAP(elemType, order=order)[:    , :    , order])}

                }
    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in face_to_cgns: elemType {elemType} is not supported')

    try:
        return faces_map[elemType % 100][face]
    except KeyError:
        raise KeyError(f'Error in face_to_cgns: face {face} is not supported')


# > Not cacheable, we pass elemNode[np.ndarray]
def dir_to_nodes(dir: str, elemType: Union[str, int], elemNodes: np.ndarray, nGeo: int) -> np.ndarray:
    """ Returns the tensor-product nodes associated with a face
    """
    if isinstance(elemType, str):
        elemType = mesh_vars.ELEMTYPE.name[elemType]

    order     = nGeo
    faces_map = {  # Tetrahedron
                   # Pyramid
                   # Wedge / Prism
                   # Hexahedron
                   8: { 'z-':              elemNodes[:    , :    , 0    ],
                        'y-': np.transpose(elemNodes[:    , 0    , :    ]),
                        'x+': np.transpose(elemNodes[order, :    , :    ]),
                        'y+':              elemNodes[:    , order, :    ],
                        'x-':              elemNodes[0    , :    , :    ],
                        'z+': np.transpose(elemNodes[:    , :    , order])}
                 }
    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in face_to_cgns: elemType {elemType} is not supported')

    try:
        return faces_map[elemType % 100][dir]
    except KeyError:
        raise KeyError(f'Error in face_to_cgns: face {dir} is not supported')


# > Not cacheable, we pass mesh[meshio.Mesh]
def count_elems(mesh: meshio.Mesh) -> int:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------
    nElems = 0
    for _, elemType in enumerate(mesh.cells_dict.keys()):
        # Only consider three-dimensional types
        if not any(s in elemType for s in mesh_vars.ELEMTYPE.type.keys()):
            continue

        ioelems = mesh.get_cells_type(elemType)
        nElems += ioelems.shape[0]
    return nElems


# > Not cacheable, we pass mesh[meshio.Mesh]
def calc_elem_bary(mesh: meshio.Mesh) -> np.ndarray:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------
    # Only consider three-dimensional types
    elem_cells = []
    for elemType in mesh.cells_dict:
        if any(s in elemType for s in mesh_vars.ELEMTYPE.type.keys()):
            elem_cells.append(mesh.get_cells_type(elemType))

    # Flatten the list of cells (concatenate all cells into one array)
    all_cells = np.concatenate(elem_cells, axis=0)

    # Calculate the centroid (mean of coordinates) for all cells at once
    return np.mean(mesh_vars.mesh.points[all_cells], axis=1)


@cache
def LINTEN(elemType: int, order: int = 1) -> np.ndarray:
    """ CGNS -> IJK ordering for element corner nodes
    """
    # Local imports ----------------------------------------
    # from pyhope.io.io_cgns import genHEXMAPCGNS
    # from pyhope.io.io_vtk import genHEXMAPVTK
    from pyhope.io.io_meshio import HEXMAPMESHIO
    # ------------------------------------------------------
    match elemType:
        # Straight-sided elements, hard-coded
        case 104:  # Tetraeder
            return np.array([0, 1, 2, 3])
        case 105:  # Pyramid
            return np.array([0, 1, 3, 2, 4])
        case 106:  # Prism
            return np.array([0, 1, 2, 3, 4, 5])
        case 108:  # Hexaeder
            return np.array([0, 1, 3, 2, 4, 5, 7, 6])
        # Curved elements, use mapping
        case 208:  # Hexaeder
            # > HEXTEN : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (1D, tensor-product style)
            # > HEXMAP : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (3D mapping)

            # # CGNS
            # _, HEXTEN = HEXMAPCGNS(order+1)

            # # VTK
            # _, HEXTEN = HEXMAPVTK(order+1)

            # MESHIO
            _, HEXTEN = HEXMAPMESHIO(order+1)
            return HEXTEN
        case _:  # Default
            print('Error in LINMAP, unknown elemType')
            sys.exit(1)


@cache
def LINMAP(elemType: int, order: int = 1) -> npt.NDArray[np.int32]:
    """ CGNS -> IJK ordering for element corner nodes
    """
    # Local imports ----------------------------------------
    # from pyhope.io.io_cgns import HEXMAPCGNS
    # from pyhope.io.io_vtk import HEXMAPVTK
    from pyhope.io.io_meshio import HEXMAPMESHIO
    # ------------------------------------------------------
    match elemType:
        # Straight-sided elements, hard-coded
        case 104:  # Tetraeder
            sys.exit(1)
            return np.array([0, 1, 2, 3])
        case 105:  # Pyramid
            sys.exit(1)
            return np.array([0, 1, 3, 2, 4])
        case 106:  # Prism
            sys.exit(1)
            return np.array([0, 1, 2, 3, 4, 5])
        case 108:  # Hexaeder
            linmap = np.zeros((2, 2, 2), dtype=int)
            indices = [ (0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0),
                        (0, 0, 1), (1, 0, 1), (1, 1, 1), (0, 1, 1) ]
            for i, index in enumerate(indices):
                linmap[index] = i
            return linmap

        # Curved elements, use mapping
        case 208:  # Hexaeder
            # > HEXTEN : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (1D, tensor-product style)
            # > HEXMAP : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (3D mapping)

            # # CGNS
            # HEXMAP, _ = HEXMAPCGNS(order+1)

            # # VTK
            # HEXMAP, _ = HEXMAPVTK(order+1)

            # MESHIO
            HEXMAP, _ = HEXMAPMESHIO(order+1)
            return HEXMAP
        case _:  # Default
            print('Error in LINMAP, unknown elemType')
            sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import os
import sys
from string import digits
from typing import cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import h5py
import meshio
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def NDOFperElemType(elemType: str, nGeo: int) -> int:
    """ Calculate the number of degrees of freedom for a given element type
    """
    match elemType:
        case 'triangle':
            return round((nGeo+1)*(nGeo+2)/2.)
        case 'quad':
            return round((nGeo+1)**2)
        case 'tetra':
            return round((nGeo+1)*(nGeo+2)*(nGeo+3)/6.)
        case 'pyramid':
            return round((nGeo+1)*(nGeo+2)*(2*nGeo+3)/6.)
        case 'prism':
            return round((nGeo+1)**(nGeo-1)*(nGeo+2)/2.)
        case 'hexahedron':
            return round((nGeo+1)**3)
        case _:
            raise ValueError(f'Unknown element type {elemType}')


def ReadHOPR(fnames: list, mesh: meshio.Mesh) -> meshio.Mesh:
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.basis.basis_basis import barycentric_weights, calc_vandermonde, change_basis_3D
    from pyhope.mesh.mesh_common import LINTEN
    from pyhope.mesh.mesh_common import faces, face_to_cgns
    from pyhope.mesh.mesh_vars import ELEMTYPE
    # ------------------------------------------------------

    hopout.sep()

    # Create an empty meshio object
    points   = mesh.points if len(mesh.points.shape)>1 else np.zeros((0, 3), dtype=np.float64)
    cells    = mesh.cells_dict
    cellsets = mesh.cell_sets

    nodeCoords   = mesh.points
    offsetnNodes = nodeCoords.shape[0]
    nSides       = 0

    # Vandermonde for changeBasis
    VdmEqHdf5ToEqMesh = np.array([])

    for fname in fnames:
        # Check if the file is using HDF5 format internally
        if not h5py.is_hdf5(fname):
            hopout.warning('[󰇘]/{} is not in HDF5 format, exiting...'.format(os.path.basename(fname)))
            sys.exit(1)

        with h5py.File(fname, mode='r') as f:
            # Check if file contains the Hopr version
            if 'HoprVersion' not in f.attrs:
                hopout.warning('[󰇘]/{} does not contain the Hopr version, exiting...'.format(os.path.basename(fname)))
                sys.exit(1)

            # Read the globalNodeIDs
            nodeInfo   = np.array(f['GlobalNodeIDs'])

            # Read the nodeCoords
            nodeCoords = np.array(f['NodeCoords'])

            # Read nGeo
            nGeo       = int(cast(int, f.attrs['Ngeo']))

            if nGeo == mesh_vars.nGeo:
                # only retain the unique nodes
                indices    = np.unique(nodeInfo, return_index=True)[1]
                nodeCoords = nodeCoords[indices]
                points     = np.append(points, nodeCoords, axis=0)
            else:
                # ChangeBasis on the non-unique nodes
                # > Currently only supported for hexahedrons
                filename = os.path.basename(fname)
                print(hopout.warn(f'[󰇘]/{filename} has different polynomial order than the current mesh, converting...',
                      length=999))
                print(hopout.warn(f'> NGeo [HDF5] = {nGeo}, NGeo [Mesh] = {mesh_vars.nGeo}') + '\n')

                # Compute the equidistant point set used by HOPR
                xEqHdf5 = np.zeros(nGeo+1)
                for i in range(nGeo+1):
                    xEqHdf5[i] = 2.*float(i)/float(nGeo) - 1.
                wBaryEqHdf5 = barycentric_weights(nGeo, xEqHdf5)

                # Compute the equidistant point set used by meshIO
                xEqMesh = np.zeros(mesh_vars.nGeo+1)
                for i in range(mesh_vars.nGeo+1):
                    xEqMesh[i] = 2.*float(i)/float(mesh_vars.nGeo) - 1.
                # wBaryEqMesh = barycentric_weights(mesh_vars.nGeo, xEqMesh)

                # Compute the Vandermonde matrix
                VdmEqHdf5ToEqMesh = calc_vandermonde(nGeo+1, mesh_vars.nGeo+1, wBaryEqHdf5, xEqHdf5, xEqMesh)

            # Read the elemInfo and sideInfo
            elemInfo   = np.array(f['ElemInfo'])
            sideInfo   = np.array(f['SideInfo'])
            BCNames    = [s.strip().decode('utf-8') for s in cast(h5py.Dataset, f['BCNames'])]

            # Construct the elements, meshio format
            for elem in elemInfo:
                # Correct ElemType if NGeo is changed
                elemNum  = elem[0] % 100
                elemNum += 200 if mesh_vars.nGeo > 1 else 100

                # Obtain the element type
                elemType = ELEMTYPE.inam[elemNum]
                if len(elemType) > 1:
                    elemType  = elemType[0].rstrip(digits)
                    elemType += str(NDOFperElemType(elemType, mesh_vars.nGeo))
                else:
                    elemType  = elemType[0]

                # ChangeBasis currently only supported for hexahedrons
                linMap    = LINTEN(elemNum, order=mesh_vars.nGeo)
                mapLin    = {k: v for v, k in enumerate(linMap)}

                if nGeo == mesh_vars.nGeo:
                    elemIDs   = np.arange(elem[4], elem[5])
                    elemNodes = np.zeros(len(elemIDs), dtype=np.uint64)
                    elemNodes = elemIDs[[mapLin[np.int64(i)] for i in range(len(elemIDs))]]
                    elemNodes = np.expand_dims(nodeInfo[elemNodes]-1+offsetnNodes, axis=0)
                else:
                    elemIDs   = np.arange(points.shape[0], points.shape[0]+(mesh_vars.nGeo+1)**3., dtype=np.uint64)
                    elemNodes = np.zeros(len(elemIDs), dtype=np.uint64)
                    elemNodes = elemIDs[[mapLin[np.int64(i)] for i in range(len(elemIDs))]]
                    # This needs no offset as we already accounted for the number of points in elemIDs
                    elemNodes = np.expand_dims(         elemNodes                , axis=0)

                    # This is still in tensor-product format
                    meshNodes = nodeCoords[np.arange(elem[4], elem[5])].reshape((nGeo+1, nGeo+1, nGeo+1, 3))
                    meshNodes = meshNodes.transpose(3, 0, 1, 2)
                    try:
                        meshNodes = change_basis_3D(VdmEqHdf5ToEqMesh, meshNodes)
                        meshNodes = meshNodes.transpose(1, 2, 3, 0)
                        meshNodes = meshNodes.reshape((int((mesh_vars.nGeo+1)**3.), 3))
                        points    = np.append(points, meshNodes, axis=0)
                    except UnboundLocalError:
                        raise UnboundLocalError('Something went wrong with the change basis')

                try:
                    cells[elemType] = np.append(cells[elemType], elemNodes.astype(np.uint64), axis=0)
                except KeyError:
                    cells[elemType] = elemNodes.astype(np.uint64)

                # Attach the boundary sides
                sCounter = 0
                for index in range(elem[2], elem[3]):
                    # Account for mortar sides
                    # TODO: Add mortar sides

                    # Obtain the side type
                    sideType  = sideInfo[index, 0]
                    sideBC    = sideInfo[index, 4]

                    BCName    = BCNames[sideBC-1]
                    face      = faces(elemType)[sCounter]
                    corners   = [elemNodes[0][s] for s in face_to_cgns(face, elemType)]

                    # Get the number of corners
                    nCorners  = abs(sideType % 10)
                    sideName  = 'quad' if nCorners == 4 else 'tri'
                    if mesh_vars.nGeo > 1:
                        sideName += str(NDOFperElemType(sideName, mesh_vars.nGeo))
                    sideNodes = np.expand_dims(corners, axis=0)

                    try:
                        cells[sideName] = np.append(cells[sideName], sideNodes.astype(np.uint64), axis=0)
                    except KeyError:
                        cells[sideName] = sideNodes.astype(np.uint64)

                    # Increment the side counter
                    sCounter += 1
                    nSides   += 1

                    if sideBC == 0:
                        continue

                    # Add the side to the cellset
                    # > We did not create any 0D/1D objects, so we do not need to consider any offset
                    try:
                        cellsets[BCName][1] = np.append(cellsets[BCName][1], np.array([nSides-1], dtype=np.uint64))
                    except KeyError:
                        # Pyright does not understand that Meshio expects a list with one None entry
                        cellsets[BCName]    = [None, np.array([nSides-1], dtype=np.uint64)]  # type: ignore

            # Update the offset for the next file
            offsetnNodes = points.shape[0]

    mesh   = meshio.Mesh(points    = points,    # noqa: E251
                         cells     = cells,     # noqa: E251
                         cell_sets = cellsets)  # noqa: E251

    return mesh



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import string
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def GenerateSides() -> None:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.common.common_progress import ProgressBar
    from pyhope.mesh.mesh_common import faces, face_to_cgns
    from pyhope.mesh.mesh_vars import ELEM, SIDE
    # ------------------------------------------------------

    hopout.sep()
    hopout.routine('Generating sides')

    mesh   = mesh_vars.mesh
    nElems = 0
    nSides = 0
    sCount = 0
    # INFO: Tuples should be faster than lists but cumbersome to add elements
    # mesh_vars.elems = ()
    # mesh_vars.sides = ()
    mesh_vars.elems = []
    mesh_vars.sides = []
    elems   = mesh_vars.elems
    sides   = mesh_vars.sides

    totalElems = 0
    for elemType in mesh.cells_dict.keys():
        # Only consider three-dimensional types
        if not any(s in elemType for s in mesh_vars.ELEMTYPE.type.keys()):
            continue

        totalElems += mesh.get_cells_type(elemType).shape[0]

    bar = ProgressBar(value=totalElems, title='│             Processing Elements', length=33)

    # Loop over all element types
    for elemType in mesh.cells_dict.keys():
        # Only consider three-dimensional types
        if not any(s in elemType for s in mesh_vars.ELEMTYPE.type.keys()):
            continue

        # Get the elements
        ioelems  = mesh.get_cells_type(elemType)
        elemMap  = mesh_vars.ELEMMAP(elemType)
        nIOElems = ioelems.shape[0]
        nIOSides = mesh_vars.ELEMTYPE.type[elemType.rstrip(string.digits)]

        # Create non-unique sides
        # mesh_vars.elems += tuple(ELEM() for _ in range(nIOElems         ))
        # mesh_vars.sides += tuple(SIDE() for _ in range(nIOElems*nIOSides))
        # elems = mesh_vars.elems
        # sides = mesh_vars.sides
        mesh_vars.elems.extend([ELEM() for _ in range(nIOElems         )])
        mesh_vars.sides.extend([SIDE() for _ in range(nIOElems*nIOSides)])

        # Create the corner faces
        corner_faces = [face_to_cgns(s, elemType) for s in faces(elemType)]

        # Create dictionaries
        for iElem in range(nElems, nElems+nIOElems):
            # elems[iElem].update(type   = elemMap,                      # noqa: E251
            #                     elemID = iElem,                        # noqa: E251
            #                     sides  = [],                           # noqa: E251
            #                     nodes  = ioelems[iElem])               # noqa: E251
            elems[iElem].type   = elemMap                       # noqa: E251
            elems[iElem].elemID = iElem                         # noqa: E251
            elems[iElem].sides  = []                            # noqa: E251
            elems[iElem].nodes  = ioelems[iElem]                # noqa: E251

            # Create the sides
            for iSide in range(nSides, nSides+nIOSides):
                # sides[iSide].update(sideType=4)
                sides[iSide].sideType=4

            # Assign corners to sides, CGNS format
            for index, face in enumerate(faces(elemType)):
                corners = [ioelems[iElem][s] for s in corner_faces[index]]
                # sides[sCount].update(face    = face,                   # noqa: E251
                #                      elemID  = iElem,                  # noqa: E251
                #                      sideID  = sCount,                 # noqa: E251
                #                      locSide = index+1,                # noqa: E251
                #                      corners = np.array(corners))      # noqa: E251
                sides[sCount].face    = face                    # noqa: E251
                sides[sCount].elemID  = iElem                   # noqa: E251
                sides[sCount].sideID  = sCount                  # noqa: E251
                sides[sCount].locSide = index+1                 # noqa: E251
                sides[sCount].corners = np.array(corners)      # noqa: E251
                sCount += 1

            # Add to nSides
            nSides += nIOSides

            # Update progress bar
            bar.step()

        # Add to nElems
        nElems += nIOElems

    # Append sides to elem
    for iSide, side in enumerate(sides):
        elemID = side.elemID
        sideID = side.sideID
        elems[elemID].sides.append(sideID)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
from scipy import spatial
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def EliminateDuplicates() -> None:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    # ------------------------------------------------------
    hopout.routine('Removing duplicate points')

    # Eliminate duplicate points
    mesh_vars.mesh.points, inverseIndices = np.unique(mesh_vars.mesh.points, axis=0, return_inverse=True)

    # Update the mesh
    for cell in mesh_vars.mesh.cells:
        # Map the old indices to the new ones
        # cell.data = np.vectorize(lambda idx: inverseIndices[idx])(cell.data)
        # Efficiently map all indices in one operation
        cell.data = inverseIndices[cell.data]

    # Also, remove near duplicate points
    # Create a KDTree for the mesh points
    points = mesh_vars.mesh.points
    tree   = spatial.KDTree(points)

    # Find all points within the tolerance
    clusters = tree.query_ball_point(points, r=mesh_vars.tolExternal)

    # Map each point to its cluster representative (first point in the cluster)
    indices = {}
    for i, cluster in enumerate(clusters):
        # Choose the minimum index as the representative for consistency
        representative = min(cluster)
        indices[i] = representative

    # Create a mapping from old indices to new indices
    indices = np.array([indices[i] for i in range(len(points))])

    # Eliminate duplicates
    _, inverseIndices = np.unique(indices, return_inverse=True)
    mesh_vars.mesh.points = points[np.unique(indices)]

    # Update the mesh cells
    for cell in mesh_vars.mesh.cells:
        cell.data = inverseIndices[cell.data]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import threading
from multiprocessing import Pool, Queue
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from alive_progress import alive_bar
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def distribute_work(elems, chunk_size: int) -> list[object]:
    """Distribute elements into chunks of a given size
    """
    return [elems[i:i + chunk_size] for i in range(0, len(elems), chunk_size)]


def update_progress(progress_queue, total_elements) -> None:
    """ Function to update the progress bar from the queue
    """
    with alive_bar(total_elements, title='│             Processing Elements', length=33) as bar:
        for _ in range(total_elements):
            # Block until we receive a progress update from the queue
            progress_queue.get()
            bar()


def run_in_parallel(process_chunk, elems, chunk_size=10):
    """Run the element processing in parallel using a specified number of processes
    """
    # Local imports ----------------------------------------
    from pyhope.common.common_vars import np_mtp
    # ------------------------------------------------------

    chunks = distribute_work(elems, chunk_size)
    total_elements = len(elems)
    progress_queue = Queue()

    # Use a separate thread for the progress bar
    progress_thread = threading.Thread(target=update_progress, args=(progress_queue, total_elements))
    progress_thread.start()

    # Use multiprocessing Pool for parallel processing
    with Pool(processes=np_mtp) as pool:
        # Map work across processes in chunks
        results = []
        for chunk_result in pool.imap_unordered(process_chunk, chunks):
            results.extend(chunk_result)
            # Update progress for each processed element in the chunk
            for _ in chunk_result:
                progress_queue.put(1)

    # Wait for the progress bar thread to finish
    progress_thread.join()
    return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
import pyhope.mesh.mesh_vars as mesh_vars
from pyhope.basis.basis_basis import change_basis_2D
from pyhope.mesh.mesh_common import face_to_nodes
# ==================================================================================================================================


def eval_nsurf(XGeo: np.ndarray, Vdm: np.ndarray, DGP: np.ndarray, wGP: np.ndarray) -> np.ndarray:
    """ Evaluate the surface integral for normals over a side of an element
    """
    # Change basis to Gauss points
    xGP      = change_basis_2D(Vdm, XGeo)

    # Compute derivatives at all Gauss points
    dXdxiGP  = np.tensordot(DGP, xGP, axes=(1, 1)).transpose(1, 0, 2)  # Shape: (3, N_GP+1, N_GP+1)
    dXdetaGP = np.tensordot(DGP, xGP, axes=(1, 2)).transpose(1, 0, 2)  # Shape: (3, N_GP+1, N_GP+1)

    # Compute the cross product at each Gauss point
    dXdxiGP  = dXdxiGP .reshape(3, -1)              # Flatten for cross computation
    dXdetaGP = dXdetaGP.reshape(3, -1)
    nVec     = np.cross(dXdxiGP, dXdetaGP, axis=0)  # Shape: (3, N_GP*N_GP)
    nVec     = nVec.reshape(3, wGP.size, wGP.size)  # Reshape to (3, N_GP+1, N_GP+1)

    # Compute the weighted normals
    weights  = np.outer(wGP, wGP)                   # Shape: (N_GP+1, N_GP+1)
    nVecW    = nVec * weights                       # Broadcast weights to shape (3, N_GP+1, N_GP+1)

    # Integrate over the Gauss points
    NSurf    = - np.sum(nVecW, axis=(1, 2))         # Sum over the last two axes
    return NSurf


def check_sides(elem,
                # points   : np.ndarray,
                VdmEqToGP: np.ndarray,
                DGP      : np.ndarray,
                wGP      : np.ndarray,
                # sides    : list
                ) -> list[bool | int | np.ndarray]:
    # Local imports ----------------------------------------
    # ------------------------------------------------------
    results = []
    points  = mesh_vars.mesh.points
    elems   = mesh_vars.elems
    sides   = mesh_vars.sides
    nGeo    = mesh_vars.nGeo

    elemType   = elem.type

    for SideID in elem.sides:
        # TODO: THIS IS CURRENTLY IGNORED, MEANING WE CHECK EVERY CONNECTION DOUBLE
        # if checked[SideID]:
        #     continue

        side   = sides[SideID]

        # Only connected sides and not small mortar sides
        # > Small mortar sides connect to big mortar side, so we will never match
        if side.connection is None or side.sideType < 0:
            continue

        # Big mortar side
        elif side.connection < 0:
            mortarType = abs(side.connection)
            nodes   = np.array([elem.nodes[s] for s in face_to_nodes(side.face, elemType, nGeo)])
            # INFO: This should be faster but I could not confirm the speedup in practice
            # nSurf   = eval_nsurf(np.moveaxis( points[  nodes], 2, 0), VdmEqToGP, DGP, wGP)
            # nSurf   = eval_nsurf(np.transpose(np.take(points,   nodes, axis=0), axes=(2, 0, 1)), VdmEqToGP, DGP, wGP)
            nSurf   = eval_nsurf(np.transpose(points[  nodes], axes=(2, 0, 1)), VdmEqToGP, DGP, wGP)
            tol     = np.linalg.norm(nSurf, ord=2) * mesh_vars.tolExternal
            # checked[SideID] = True

            # Mortar sides are the following virtual sides
            nMortar = 4 if mortarType == 1 else 2
            nnbSurf = np.zeros((3,), dtype=float)
            for mortarSide in range(nMortar):
                # Get the matching side
                nbside   = sides[sides[SideID + mortarSide + 1].connection]
                nbelem   = elems[nbside.elemID]
                nbnodes  = np.array([nbelem.nodes[s] for s in face_to_nodes(nbside.face, nbelem.type, nGeo)])
                # INFO: This should be faster but I could not confirm the speedup in practice
                # nnbSurf += eval_nsurf(np.moveaxis(points[nbnodes], 2, 0), VdmEqToGP, DGP, wGP)
                nnbSurf += eval_nsurf(np.transpose(points[nbnodes], axes=(2, 0, 1)), VdmEqToGP, DGP, wGP)
                # checked[nbside] = True

            # Check if side normals are within tolerance
            nSurfErr = np.sum(np.abs(nnbSurf + nSurf))
            success  = nSurfErr < tol

        # Internal side
        elif side.connection >= 0:
            # Ignore the virtual mortar sides
            if side.locMortar is not None:
                continue

            nodes   = np.array([elem.nodes[s] for s in face_to_nodes(  side.face,   elemType, nGeo)])
            # INFO: This should be faster but I could not confirm the speedup in practice
            # nSurf   = eval_nsurf(np.moveaxis( points[  nodes], 2, 0), VdmEqToGP, DGP, wGP)
            nSurf   = eval_nsurf(np.transpose(points[  nodes], axes=(2, 0, 1)), VdmEqToGP, DGP, wGP)
            tol     = np.linalg.norm(nSurf, ord=2) * mesh_vars.tolExternal
            # checked[SideID] = True

            # Connected side
            nbside  = sides[side.connection]
            nbelem  = elems[nbside.elemID]
            nbnodes = np.array([nbelem.nodes[s] for s in face_to_nodes(nbside.face, nbelem.type, nGeo)])
            # INFO: This should be faster but I could not confirm the speedup in practice
            # nnbSurf = eval_nsurf(np.moveaxis(points[nbnodes], 2, 0), VdmEqToGP, DGP, wGP)
            nnbSurf = eval_nsurf(np.transpose(points[nbnodes], axes=(2, 0, 1)), VdmEqToGP, DGP, wGP)
            # checked[nbside] = True

            # Check if side normals are within tolerance
            nSurfErr = np.sum(np.abs(nnbSurf + nSurf))
            success  = nSurfErr < tol

        else:
            continue

        results.append((success, SideID, nSurf, nnbSurf, nSurfErr, tol))
    return results


def process_chunk(chunk) -> np.ndarray:
    """Process a chunk of elements by checking surface normal orientation
    """
    chunk_results    = np.empty(len(chunk), dtype=object)
    # elem, VdmEqToGP, DGP, wGP = elem_data
    chunk_results[:] = [check_sides(*elem_data) for elem_data in chunk]
    return chunk_results


def CheckWatertight() -> None:
    """ Check if the mesh is watertight
    """
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.basis.basis_basis import barycentric_weights, legendre_gauss_nodes
    from pyhope.basis.basis_basis import calc_vandermonde, polynomial_derivative_matrix
    from pyhope.common.common_parallel import run_in_parallel
    from pyhope.common.common_vars import np_mtp
    from pyhope.readintools.readintools import GetLogical
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('CHECK WATERTIGHTNESS...')
    hopout.sep()

    checkWatertightness = GetLogical('CheckWatertightness')
    if not checkWatertightness:
        return None

    nGeo = mesh_vars.nGeo + 1

    # Compute the equidistant point set used by meshIO
    xEq = np.zeros(nGeo)
    for i in range(nGeo):
        xEq[i] = 2.*float(i)/float(nGeo-1) - 1.
    wBaryEq   = barycentric_weights(nGeo, xEq)

    xGP, wGP  = legendre_gauss_nodes(nGeo)
    DGP       = polynomial_derivative_matrix(nGeo, xGP)
    VdmEqToGP = calc_vandermonde(nGeo, nGeo, wBaryEq, xEq, xGP)

    # Check all sides
    elems     = mesh_vars.elems
    sides     = mesh_vars.sides
    points    = mesh_vars.mesh.points
    # checked = np.zeros((len(sides)), dtype=bool)

    # Prepare elements for parallel processing
    if np_mtp > 0:
        tasks  = [(elem, VdmEqToGP, DGP, wGP)
                  for elem in elems]
        # Run in parallel with a chunk size
        # > Dispatch the tasks to the workers, minimum 10 tasks per worker, maximum 1000 tasks per worker
        res    = run_in_parallel(process_chunk, tasks, chunk_size=max(1, min(1000, max(10, int(len(tasks)/(200.*np_mtp))))))
    else:
        res    = np.empty(len(elems), dtype=object)
        res[:] = [check_sides(elem, VdmEqToGP, DGP, wGP) for elem in elems]

    for r in res:
        for result in r:
            if bool(result[0]) is not True:
                # Unpack the results
                side     = sides[result[1]]
                elem     = elems[side.elemID]
                nbside   = side.connection

                nSurf    = result[2]
                nnbSurf  = result[3]

                nSurfErr = result[4]
                tol      = result[5]

                nodes    = np.transpose(np.array([elem.nodes[s] for s in face_to_nodes(side.face, elem.type, mesh_vars.nGeo)]))
                nodes    = np.transpose(points[nodes]         , axes=(2, 0, 1))
                nbelem   = elems[sides[nbside].elemID]
                nbnodes  = np.transpose(np.array([nbelem.nodes[s] for s in face_to_nodes(side.face, nbelem.type, mesh_vars.nGeo)]))
                nnbodes  = np.transpose(points[nbnodes], axes=(2, 0, 1))

                strLen = max(len(str(side.sideID+1)), len(str(nbside)))
                hopout.warning('Watertightness check failed!')
                print(hopout.warn(f'> Element {elem.elemID+1:>{strLen}}, Side {side.face}, Side {side.sideID+1:>{strLen}}'))  # noqa: E501
                print(hopout.warn('> Normal vector: [' + ' '.join('{:12.3f}'.format(s) for s in nSurf) + ']'))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nodes[:,  0,  0]) + ']'))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nodes[:,  0, -1]) + ']'))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nodes[:, -1,  0]) + ']'))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nodes[:, -1, -1]) + ']'))
                print()
                print(hopout.warn(f'> Element {sides[nbside].elemID+1:>{strLen}}, Side {sides[nbside].face}, Side {nbside+1:>{strLen}}'))  # noqa: E501
                print(hopout.warn('> Normal vector: [' + ' '.join('{:12.3f}'.format(s) for s in nnbSurf) + ']'))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nnbodes[:,  0,  0]) + ']'))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nnbodes[:,  0, -1]) + ']'))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nnbodes[:, -1,  0]) + ']'))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in nnbodes[:, -1, -1]) + ']'))

                # Check if side is oriented inwards
                if nSurfErr < 0:
                    hopout.warning('Side is oriented inwards, exiting...')
                else:
                    hopout.warning(f'Surface normals are not within tolerance {nSurfErr} > {tol}, exiting...')
                sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import plotext as plt
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
from pyhope.basis.basis_basis import evaluate_jacobian
# ==================================================================================================================================


def plot_histogram(data: np.ndarray) -> None:
    """ Plot a histogram of all Jacobians
    """
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    from pyhope.output.output import STD_LENGTH
    # ------------------------------------------------------

    ticks = ['│<0.0      │',
             '│ 0.0-0.1  │',
             '│ 0.1-0.2  │',
             '│ 0.2-0.3  │',
             '│ 0.3-0.4  │',
             '│ 0.4-0.5  │',
             '│ 0.5-0.6  │',
             '│ 0.6-0.7  │',
             '│ 0.7-0.8  │',
             '│ 0.8-0.9  │',
             # '│ 0.9-0.99 │',
             '│>0.9-1.0  │']

    # Define the bins for categorizing jacobians
    # bins     = [ -np.inf, 0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.99, np.inf]
    bins     = [ -np.inf, 0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9,       np.inf]

    # Use np.histogram to count jacobians in the defined bins
    count, _ = np.histogram(data, bins=bins)

    # Setup plot
    hopout.sep()
    hopout.info('Scaled Jacobians')
    hopout.separator(18)
    plt.simple_bar(ticks, count, width=STD_LENGTH)
    plt.show()
    hopout.separator(18)


def process_chunk(chunk) -> list[np.ndarray]:
    """Process a chunk of elements by evaluating the Jacobian for each
    """
    chunk_results = []
    for elem in chunk:
        # nodeCoords, nGeoRef, VdmGLtoAP, D_EqToGL = elem
        nodeCoords, _, VdmGLtoAP, D_EqToGL = elem
        jac    = evaluate_jacobian(nodeCoords, VdmGLtoAP, D_EqToGL)
        # INFO: ALTERNATIVE VERSION, CACHING VDM, D
        # nodeCoords, evaluate_jacobian = elem
        # jac    = evaluate_jacobian(nodeCoords)
        maxJac = np.max(np.abs(jac))
        minJac = np.min(jac)
        chunk_results.append(minJac / maxJac)
    return chunk_results


def CheckJacobians() -> None:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.basis.basis_basis import barycentric_weights, polynomial_derivative_matrix, calc_vandermonde
    from pyhope.basis.basis_basis import legendre_gauss_lobatto_nodes
    # INFO: ALTERNATIVE VERSION, CACHING VDM, D
    # from pyhope.basis.basis_basis import JacobianEvaluator
    from pyhope.common.common_parallel import run_in_parallel
    from pyhope.common.common_vars import np_mtp
    from pyhope.readintools.readintools import GetLogical
    from pyhope.mesh.mesh_common import LINTEN
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('CHECK JACOBIANS...')
    hopout.sep()

    checkElemJacobians = GetLogical('CheckElemJacobians')
    if not checkElemJacobians:
        return None

    nGeo = mesh_vars.nGeo + 1

    # Compute the equidistant point set used by meshIO
    xEq = np.zeros(nGeo)
    for i in range(nGeo):
        xEq[i] = 2.*float(i)/float(nGeo-1) - 1.
    wBaryEq   = barycentric_weights(nGeo, xEq)

    xGL, _    = legendre_gauss_lobatto_nodes(nGeo)
    DGL       = polynomial_derivative_matrix(nGeo, xGL)
    VdmEqToGL = calc_vandermonde(nGeo, nGeo, wBaryEq, xEq, xGL)
    wbaryGL   = barycentric_weights(nGeo, xGL)

    D_EqToGL  = np.matmul(DGL, VdmEqToGL)

    # Interpolate derivatives on GL (N) to nGeoRef points
    nGeoRef = 3*(nGeo-1)+1
    xAP     = np.zeros(nGeoRef)
    for i in range(nGeoRef):
        xAP[i] = 2. * float(i)/float(nGeoRef-1) - 1.
    VdmGLtoAP = calc_vandermonde(nGeo, nGeoRef, wbaryGL, xGL, xAP)
    # INFO: ALTERNATIVE VERSION, CACHING VDM, D
    # evaluate_jacobian = JacobianEvaluator(VdmGLtoAP, D_EqToGL).evaluate_jacobian

    # Map all points to tensor product
    elems = mesh_vars.elems
    nodes = mesh_vars.mesh.points

    # Prepare elements for parallel processing
    tasks = []

    for elem in elems:
        # Only consider hexahedrons
        if int(elem.type) % 100 != 8:
            continue

        # Get the mapping
        linMap = LINTEN(elem.type, order=mesh_vars.nGeo)
        mapLin = {k: v for v, k in enumerate(linMap)}

        # Fill the NodeCoords
        nodeCoords = np.zeros((nGeo ** 3, 3), dtype=np.float64)
        nodeCoords[[mapLin[np.int64(iNode)] for iNode in range(len(elem.nodes))]] = nodes[elem.nodes]

        xGeo = np.zeros((3, nGeo, nGeo, nGeo))
        xGeo = nodeCoords[:nGeo**3].reshape((nGeo, nGeo, nGeo, 3), order='F').transpose(3, 0, 1, 2)

        if np_mtp > 0:
            # Add tasks for parallel processing
            tasks.append((xGeo, nGeoRef, VdmGLtoAP, D_EqToGL))
            # INFO: ALTERNATIVE VERSION, CACHING VDM, D
            # tasks.append((xGeo, evaluate_jacobian))
        else:
            jac = evaluate_jacobian(xGeo, VdmGLtoAP, D_EqToGL)
            # INFO: ALTERNATIVE VERSION, CACHING VDM, D
            # jac = evaluate_jacobian(xGeo)
            maxJac =  np.max(np.abs(jac))
            minJac =  np.min(       jac)
            tasks.append(minJac / maxJac)

    if np_mtp > 0:
        # Run in parallel with a chunk size
        # > Dispatch the tasks to the workers, minimum 10 tasks per worker, maximum 1000 tasks per worker
        jacs = run_in_parallel(process_chunk, tasks, chunk_size=max(1, min(1000, max(10, int(len(tasks)/(200.*np_mtp))))))
    else:
        jacs = np.array(tasks)

    # Plot the histogram of the Jacobians
    plot_histogram(np.array(jacs))
